----------------------------------------------------------------------------------
-- COPYRIGHT 2019 Victor Ral Arteaga Prez
--This program is free software: you can redistribute it and/or modify
--it under the terms of the GNU General Public License as published by
--the Free Software Foundation, either version 3 of the License, or
--(at your option) any later version.
--
--This program is distributed in the hope that it will be useful,
--but WITHOUT ANY WARRANTY; without even the implied warranty of
--MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--GNU General Public License for more details.
--
--You should have received a copy of the GNU General Public License
--along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--
-- Description: Librera que controla el DAC mediante interfaz SPI
--              Adquiere el valor digital y lo manda por el protocolo serial
--              siguiendo las especificaciones del integrado MCP4822. 
--              Las caracteristicas de este modulo son:
--					
--              * DAC de dos canales con 12 bits de resolucin
--              * Genera protocolo SPI a 16.6MHz
--              * La bandera CONV_FIN indica el fin de la conversin
--
-- Para iniciar una conversin se debe poner el reset en bajo y mantener el dato de la conversin
-- hasta que la conversin finalice. Tal y como lo muestra el siguiente diagrama de tiempo:      
--
--			RESET 	\________________________________________________/||
--			CH       __________________________________________________________________________||
--			DATADAC  ----XXX----><--------------------X"00A"---------------------><----XXX-----||            
--			CONV_FIN ________________________________________________________/\_____________||
--                   ------------|----------------1.08 us--------------------|--------------||
--  
------------------------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity LIB_DAC_AVANXE7_REVA is

Port ( 
    CLK      : IN STD_LOGIC;	-- Reloj FPGA a 100MHz
    RESET    : IN STD_LOGIC;	-- Reset del DAC
    CH       : IN STD_LOGIC;	-- Selecciona uno de los dos canales
    DATA_DAC : IN STD_LOGIC_VECTOR(11 DOWNTO 0); -- Valor digital a convertir de 0 a 4095
    SDI      : OUT STD_LOGIC;	-- Puerto serial para SPI
    CS       : OUT STD_LOGIC;	-- Chip Select para SPI
    SCK      : OUT STD_LOGIC;	-- Reloj para SPI
    LDAC     : OUT STD_LOGIC;	-- Pin LDAC de MCP4822
    CONV_FIN : OUT STD_LOGIC	-- Bit que indica fin de conversin y slo dura un ciclo de reloj
);

end LIB_DAC_AVANXE7_REVA;

architecture Behavioral of LIB_DAC_AVANXE7_REVA is

-- Seales para mquina de estados que genera el protocolo SPI
type state is (idle,tx_rx,fin);
signal pr_state   : state;            -- Estado presente de mquina de estados
signal nx_state   : state;            -- Estado futuro de mquina de estados
signal clk_ms     : std_logic := '0'; -- Seal para protocolo SPI
signal buffer_out : std_logic_vector(15 downto 0) := (others => '0'); -- Buffer con toda la informacin a transmitir
signal i    		: integer range 0 to 31 := 0; -- Selecciona el bit a enviar
signal bits 		: integer range 0 to 31 := 1; -- Indica el nmero de bits que se transmitirn

-- Seales de mquina de estados que genera bandera CONV_FIN
signal edook: integer range 0 to 1 := 0;  
signal ok: std_logic := '0';

-- Registro que almacena temporalmete al dato a convertir
signal data_dacs : std_logic_vector(11 downto 0); 

begin

-- Se cargan los datos en el buffer
buffer_out <= CH&"011" & DATA_DACs;

-- Proceso que genera la seal de reloj para SPI
process(clk)
variable contador: integer range 0 to 65535;
begin
if rising_edge(clk) then
    contador := contador + 1;
    if(contador=3) then
        clk_ms<=not clk_ms;
        contador := 0;
    end if;
end if;
end process;

-- Proceso que controla la mquina de estados. Dependiendo el nmero de bits transmitidos hace la transicin de estado. 
process(clk_ms,reset)
begin
if reset = '1' then -- Si el reset est activado se para la transmisin 
    pr_state <= idle;
    i <= 0;
elsif rising_edge(clk_ms) then
	if (i = bits - 1) then	-- Si el nmero de bis lleg a su valor mximo entonces pasa al siguiente estado	
		pr_state <= nx_state;
		i <= 0;
		ok <= '1'; -- Se activa "ok" para la mquina de estados de CONV_FIN
	else
	   data_dacs <= DATA_DAC;
		i <= i + 1;
		ok <= '0';
	end if;
end if;
END PROCESS;

-- Mquina de estados que genera protocolo SPI           
process(pr_state,clk_ms,i,buffer_out)
begin

 case pr_state is
 
  when idle => --estado neutral 
	  SCK <= '1';
	  SDI <= '0';
	  LDAC <= '1';
	  CS <= '0';
	  bits <= 1;	  
	  nx_state <= tx_rx;
 
	when tx_rx => -- Estado de transmisin 
	  SCK <= not clk_ms;
	  SDI <= buffer_out(15 - i);
	  LDAC <= '1';
      CS <= '0';
	  bits <= 16;
	  nx_state <= fin;
	  
	when fin => -- Estado fin de transmisin
	  SCK <= '1';
	  SDI <= '0';
	  LDAC <= '0';
      CS <= '1';
	  bits <= 1;	 
	  nx_state <= idle;
  
 end case;
end process;

-- Mquina de estados que genera seal de CONV_FIN
PROCESS(CLK)
BEGIN
    IF RISING_EDGE(CLK) THEN
        IF edook = 0 THEN
            IF ok = '1' THEN
                edook <= 1;
                CONV_FIN <= '1';
            ELSE
                CONV_FIN <= '0';
            END IF;
        ELSE 
            IF ok = '0' THEN
                edook <= 0;
            ELSE
                edook <= 1;
            END IF;
            CONV_FIN <= '0';
        END IF;
    END IF;
END PROCESS;


end Behavioral;

